# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tubefetch.core.pipeline."""

from datetime import datetime, timezone
from pathlib import Path
from unittest.mock import patch

from tubefetch.core.errors import FetchPhase
from tubefetch.core.models import FetchResult, Metadata, Transcript, TranscriptSegment
from tubefetch.core.options import FetchOptions
from tubefetch.core.pipeline import process_video
from tubefetch.services.media import MediaResult
from tubefetch.services.metadata import MetadataError
from tubefetch.services.transcript import TranscriptError, TranscriptNotFound


def _make_metadata(video_id: str = "dQw4w9WgXcQ") -> Metadata:
    return Metadata(
        video_id=video_id,
        source_url=f"https://www.youtube.com/watch?v={video_id}",
        title="Test Video",
        fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
        metadata_source="yt-dlp",
    )


def _make_transcript(video_id: str = "dQw4w9WgXcQ") -> Transcript:
    return Transcript(
        video_id=video_id,
        language="en",
        is_generated=False,
        segments=[TranscriptSegment(start=0.0, duration=2.0, text="Hello")],
        fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
        transcript_source="youtube-transcript-api",
    )


class TestProcessVideo:
    """Test the per-video pipeline."""

    @patch("tubefetch.core.pipeline.download_media")
    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_success_full_pipeline(self, mock_meta, mock_trans, mock_media, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        mock_media.return_value = MediaResult(video_id="dQw4w9WgXcQ", skipped=True)

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        assert isinstance(result, FetchResult)
        assert result.success is True
        assert result.video_id == "dQw4w9WgXcQ"
        assert result.metadata_path is not None
        assert result.transcript_path is not None
        assert result.metadata_path.exists()
        assert result.transcript_path.exists()
        assert result.errors == []
        mock_meta.assert_called_once()
        mock_trans.assert_called_once()

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_creates_output_dir(self, mock_meta, mock_trans, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()

        opts = FetchOptions(out=tmp_path)
        process_video("dQw4w9WgXcQ", opts)

        assert (tmp_path / "dQw4w9WgXcQ").is_dir()

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_metadata_error_continues(self, mock_meta, mock_trans, tmp_path):
        from tubefetch.core.errors import FetchErrorCode

        mock_meta.side_effect = MetadataError("Video not found", code=FetchErrorCode.VIDEO_NOT_FOUND)
        mock_trans.return_value = _make_transcript()

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        assert result.success is False
        assert result.metadata_path is None
        assert result.transcript_path is not None
        assert len(result.errors) == 1
        assert result.errors[0].phase == FetchPhase.METADATA

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_transcript_error_continues(self, mock_meta, mock_trans, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.side_effect = TranscriptNotFound("No transcript")

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        assert result.success is True
        assert result.metadata_path is not None
        assert result.transcript_path is None
        assert len(result.errors) == 1
        assert result.errors[0].phase == FetchPhase.TRANSCRIPT

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_both_errors(self, mock_meta, mock_trans, tmp_path):
        from tubefetch.core.errors import FetchErrorCode

        mock_meta.side_effect = MetadataError("fail", code=FetchErrorCode.VIDEO_NOT_FOUND)
        mock_trans.side_effect = TranscriptError("fail", code=FetchErrorCode.TRANSCRIPT_NOT_FOUND)

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        assert result.success is False
        assert len(result.errors) == 2

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_cache_skip_metadata(self, mock_meta, mock_trans, tmp_path):
        # Pre-create cached metadata file with valid data
        meta = _make_metadata()
        video_dir = tmp_path / "dQw4w9WgXcQ"
        video_dir.mkdir()
        import json

        (video_dir / "metadata.json").write_text(json.dumps(meta.model_dump(mode="json"), default=str))

        mock_trans.return_value = _make_transcript()

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        mock_meta.assert_not_called()
        assert result.metadata_path == video_dir / "metadata.json"
        assert result.metadata is not None
        assert result.metadata.video_id == "dQw4w9WgXcQ"
        assert result.metadata.title == "Test Video"

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_cache_skip_transcript(self, mock_meta, mock_trans, tmp_path):
        # Pre-create cached transcript file with valid data
        trans = _make_transcript()
        video_dir = tmp_path / "dQw4w9WgXcQ"
        video_dir.mkdir()
        import json

        (video_dir / "transcript.json").write_text(json.dumps(trans.model_dump(mode="json"), default=str))

        mock_meta.return_value = _make_metadata()

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        mock_trans.assert_not_called()
        assert result.transcript_path == video_dir / "transcript.json"
        assert result.transcript is not None
        assert result.transcript.video_id == "dQw4w9WgXcQ"
        assert result.transcript.language == "en"
        assert len(result.transcript.segments) == 1

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_force_overrides_cache(self, mock_meta, mock_trans, tmp_path):
        # Pre-create cached files
        video_dir = tmp_path / "dQw4w9WgXcQ"
        video_dir.mkdir()
        (video_dir / "metadata.json").write_text("{}")
        (video_dir / "transcript.json").write_text("{}")

        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()

        opts = FetchOptions(out=tmp_path, force=True)
        result = process_video("dQw4w9WgXcQ", opts)

        mock_meta.assert_called_once()
        mock_trans.assert_called_once()
        assert result.success is True

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_force_metadata_selective(self, mock_meta, mock_trans, tmp_path):
        video_dir = tmp_path / "dQw4w9WgXcQ"
        video_dir.mkdir()
        (video_dir / "metadata.json").write_text("{}")
        (video_dir / "transcript.json").write_text("{}")

        mock_meta.return_value = _make_metadata()

        opts = FetchOptions(out=tmp_path, force_metadata=True)
        process_video("dQw4w9WgXcQ", opts)

        mock_meta.assert_called_once()
        mock_trans.assert_not_called()

    @patch("tubefetch.core.pipeline.download_media")
    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_media_download_when_enabled(self, mock_meta, mock_trans, mock_media, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        mock_media.return_value = MediaResult(
            video_id="dQw4w9WgXcQ",
            paths=[Path("/tmp/video.mp4")],
        )

        opts = FetchOptions(out=tmp_path, download="video")
        result = process_video("dQw4w9WgXcQ", opts)

        assert result.success is True
        assert len(result.media_paths) == 1
        mock_media.assert_called_once()

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_no_media_when_download_none(self, mock_meta, mock_trans, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()

        opts = FetchOptions(out=tmp_path, download="none")
        result = process_video("dQw4w9WgXcQ", opts)

        assert result.media_paths == []

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_returns_metadata_and_transcript_objects(self, mock_meta, mock_trans, tmp_path):
        meta = _make_metadata()
        trans = _make_transcript()
        mock_meta.return_value = meta
        mock_trans.return_value = trans

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        assert result.metadata == meta
        assert result.transcript == trans

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_success_true_when_only_transcript_fails(self, mock_meta, mock_trans, tmp_path):
        mock_meta.return_value = _make_metadata()
        from tubefetch.core.errors import FetchErrorCode

        mock_trans.side_effect = TranscriptError("No captions", code=FetchErrorCode.TRANSCRIPT_NOT_FOUND)

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        assert result.success is True
        assert result.metadata is not None
        assert result.transcript is None
        assert len(result.errors) == 1
        assert result.errors[0].phase == FetchPhase.TRANSCRIPT

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_success_false_when_metadata_fails(self, mock_meta, mock_trans, tmp_path):
        from tubefetch.core.errors import FetchErrorCode

        mock_meta.side_effect = MetadataError("Video not found", code=FetchErrorCode.VIDEO_NOT_FOUND)
        mock_trans.return_value = _make_transcript()

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        assert result.success is False
        assert result.metadata is None
        assert result.transcript is not None
        assert any(e.phase == FetchPhase.METADATA for e in result.errors)

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_cached_rerun_populates_both_objects(self, mock_meta, mock_trans, tmp_path):
        """Simulate a second run where both files are cached on disk."""
        import json

        meta = _make_metadata()
        trans = _make_transcript()
        video_dir = tmp_path / "dQw4w9WgXcQ"
        video_dir.mkdir()
        (video_dir / "metadata.json").write_text(json.dumps(meta.model_dump(mode="json"), default=str))
        (video_dir / "transcript.json").write_text(json.dumps(trans.model_dump(mode="json"), default=str))

        opts = FetchOptions(out=tmp_path)
        result = process_video("dQw4w9WgXcQ", opts)

        mock_meta.assert_not_called()
        mock_trans.assert_not_called()
        assert result.success is True
        assert result.metadata is not None
        assert result.metadata.video_id == "dQw4w9WgXcQ"
        assert result.transcript is not None
        assert result.transcript.language == "en"
        assert len(result.transcript.segments) == 1



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tubefetch.services.transcript."""

from dataclasses import dataclass
from unittest.mock import MagicMock, patch

import pytest

from tubefetch.core.models import Transcript
from tubefetch.core.options import FetchOptions
from tubefetch.services.transcript import (
    TranscriptError,
    TranscriptNotFound,
    _select_transcript,
    get_transcript,
    list_available_transcripts,
)

# --- Helpers for mocking youtube-transcript-api objects ---


@dataclass
class FakeTranscriptEntry:
    """Mimics a Transcript object from youtube-transcript-api's TranscriptList."""

    language_code: str
    language: str
    is_generated: bool

    def fetch(self):
        return FakeFetchedTranscript(
            video_id="test_vid",
            language_code=self.language_code,
            is_generated=self.is_generated,
            snippets=[
                FakeSnippet(text="Hello", start=0.0, duration=2.0),
                FakeSnippet(text="World", start=2.0, duration=3.0),
            ],
        )


@dataclass
class FakeSnippet:
    text: str
    start: float
    duration: float


@dataclass
class FakeFetchedTranscript:
    video_id: str
    language_code: str
    is_generated: bool
    snippets: list

    def __iter__(self):
        return iter(self.snippets)


# --- _select_transcript ---


class TestSelectTranscript:
    """Test the language selection algorithm."""

    def _make_entries(self, specs):
        """Create FakeTranscriptEntry list from (lang, is_generated) tuples."""
        return [
            FakeTranscriptEntry(
                language_code=lang,
                language=lang,
                is_generated=gen,
            )
            for lang, gen in specs
        ]

    def test_prefers_manual_in_preferred_language(self):
        available = self._make_entries([("en", False), ("en", True), ("es", False)])
        result = _select_transcript(available, languages=["en"], allow_generated=True, allow_any_language=False)
        assert result.language_code == "en"
        assert result.is_generated is False

    def test_falls_back_to_generated_when_allowed(self):
        available = self._make_entries([("en", True), ("es", False)])
        result = _select_transcript(available, languages=["en"], allow_generated=True, allow_any_language=False)
        assert result.language_code == "en"
        assert result.is_generated is True

    def test_skips_generated_when_not_allowed(self):
        available = self._make_entries([("en", True), ("es", False)])
        result = _select_transcript(available, languages=["en"], allow_generated=False, allow_any_language=False)
        assert result is None

    def test_language_priority_order(self):
        available = self._make_entries([("es", False), ("en", False)])
        result = _select_transcript(available, languages=["en", "es"], allow_generated=True, allow_any_language=False)
        assert result.language_code == "en"

    def test_second_language_fallback(self):
        available = self._make_entries([("es", False), ("de", False)])
        result = _select_transcript(available, languages=["en", "es"], allow_generated=True, allow_any_language=False)
        assert result.language_code == "es"

    def test_allow_any_language_manual_first(self):
        available = self._make_entries([("ja", False), ("ja", True)])
        result = _select_transcript(available, languages=["en"], allow_generated=True, allow_any_language=True)
        assert result.language_code == "ja"
        assert result.is_generated is False

    def test_allow_any_language_generated_fallback(self):
        available = self._make_entries([("ja", True)])
        result = _select_transcript(available, languages=["en"], allow_generated=True, allow_any_language=True)
        assert result.language_code == "ja"
        assert result.is_generated is True

    def test_allow_any_language_no_generated(self):
        available = self._make_entries([("ja", True)])
        result = _select_transcript(available, languages=["en"], allow_generated=False, allow_any_language=True)
        assert result is None

    def test_no_transcripts_available(self):
        result = _select_transcript([], languages=["en"], allow_generated=True, allow_any_language=True)
        assert result is None

    def test_preferred_language_beats_any_language(self):
        available = self._make_entries([("ja", False), ("en", True)])
        result = _select_transcript(available, languages=["en"], allow_generated=True, allow_any_language=True)
        assert result.language_code == "en"


# --- get_transcript ---


class TestGetTranscript:
    """Test get_transcript with mocked youtube-transcript-api."""

    @patch("tubefetch.services.transcript.YouTubeTranscriptApi")
    def test_success(self, mock_api_class):
        mock_api = MagicMock()
        mock_api_class.return_value = mock_api

        entries = [
            FakeTranscriptEntry(language_code="en", language="English", is_generated=False),
        ]
        mock_api.list.return_value = iter(entries)

        options = FetchOptions(languages=["en"])
        result = get_transcript("dQw4w9WgXcQ", options)

        assert isinstance(result, Transcript)
        assert result.video_id == "dQw4w9WgXcQ"
        assert result.language == "en"
        assert result.is_generated is False
        assert len(result.segments) == 2
        assert result.segments[0].text == "Hello"
        assert result.transcript_source == "youtube-transcript-api"
        assert result.available_languages == ["en"]

    @patch("tubefetch.services.transcript.YouTubeTranscriptApi")
    def test_transcript_not_found(self, mock_api_class):
        mock_api = MagicMock()
        mock_api_class.return_value = mock_api
        mock_api.list.return_value = iter([])

        options = FetchOptions(languages=["en"])
        with pytest.raises(TranscriptNotFound, match="TRANSCRIPT_NOT_FOUND"):
            get_transcript("xxxxxxxxxxx", options)

    @patch("tubefetch.services.transcript.YouTubeTranscriptApi")
    def test_transcripts_disabled(self, mock_api_class):
        from youtube_transcript_api import TranscriptsDisabled

        mock_api = MagicMock()
        mock_api_class.return_value = mock_api
        mock_api.list.side_effect = TranscriptsDisabled("xxxxxxxxxxx")

        options = FetchOptions(languages=["en"])
        with pytest.raises(TranscriptError, match="Transcripts are disabled"):
            get_transcript("xxxxxxxxxxx", options)

    @patch("tubefetch.services.transcript.YouTubeTranscriptApi")
    def test_generated_transcript_when_allowed(self, mock_api_class):
        mock_api = MagicMock()
        mock_api_class.return_value = mock_api

        entries = [
            FakeTranscriptEntry(language_code="en", language="English (auto)", is_generated=True),
        ]
        mock_api.list.return_value = iter(entries)

        options = FetchOptions(languages=["en"], allow_generated=True)
        result = get_transcript("dQw4w9WgXcQ", options)
        assert result.is_generated is True

    @patch("tubefetch.services.transcript.YouTubeTranscriptApi")
    def test_generated_rejected_when_not_allowed(self, mock_api_class):
        mock_api = MagicMock()
        mock_api_class.return_value = mock_api

        entries = [
            FakeTranscriptEntry(language_code="en", language="English (auto)", is_generated=True),
        ]
        mock_api.list.return_value = iter(entries)

        options = FetchOptions(languages=["en"], allow_generated=False)
        with pytest.raises(TranscriptNotFound):
            get_transcript("dQw4w9WgXcQ", options)

    @patch("tubefetch.services.transcript.YouTubeTranscriptApi")
    def test_any_language_fallback(self, mock_api_class):
        mock_api = MagicMock()
        mock_api_class.return_value = mock_api

        entries = [
            FakeTranscriptEntry(language_code="ja", language="Japanese", is_generated=False),
        ]
        mock_api.list.return_value = iter(entries)

        options = FetchOptions(languages=["en"], allow_any_language=True)
        result = get_transcript("dQw4w9WgXcQ", options)
        assert result.language == "ja"


# --- list_available_transcripts ---


class TestListAvailableTranscripts:
    """Test list_available_transcripts."""

    @patch("tubefetch.services.transcript.YouTubeTranscriptApi")
    def test_lists_all(self, mock_api_class):
        mock_api = MagicMock()
        mock_api_class.return_value = mock_api

        entries = [
            FakeTranscriptEntry(language_code="en", language="English", is_generated=False),
            FakeTranscriptEntry(language_code="en", language="English (auto)", is_generated=True),
            FakeTranscriptEntry(language_code="es", language="Spanish", is_generated=False),
        ]
        mock_api.list.return_value = iter(entries)

        result = list_available_transcripts("dQw4w9WgXcQ")
        assert len(result) == 3
        assert result[0] == {"language_code": "en", "language": "English", "is_generated": False}
        assert result[1]["is_generated"] is True
        assert result[2]["language_code"] == "es"

    @patch("tubefetch.services.transcript.YouTubeTranscriptApi")
    def test_disabled(self, mock_api_class):
        from youtube_transcript_api import TranscriptsDisabled

        mock_api = MagicMock()
        mock_api_class.return_value = mock_api
        mock_api.list.side_effect = TranscriptsDisabled("xxxxxxxxxxx")

        with pytest.raises(TranscriptError, match="Transcripts are disabled"):
            list_available_transcripts("xxxxxxxxxxx")



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Token bucket rate limiter."""

from __future__ import annotations

import threading
import time


class TokenBucket:
    """Thread-safe token bucket rate limiter.

    Args:
        rate: Tokens added per second (e.g. 2.0 = 2 requests per second).
        capacity: Maximum burst capacity. Defaults to rate (no burst beyond 1s).
    """

    def __init__(self, rate: float = 2.0, capacity: float | None = None) -> None:
        self._rate = rate
        self._capacity = capacity if capacity is not None else rate
        self._tokens = self._capacity
        self._last_refill = time.monotonic()
        self._lock = threading.Lock()

    @property
    def rate(self) -> float:
        return self._rate

    @property
    def capacity(self) -> float:
        return self._capacity

    def acquire(self, tokens: float = 1.0, blocking: bool = True) -> bool:
        """Acquire tokens from the bucket.

        Args:
            tokens: Number of tokens to consume (default 1).
            blocking: If True, block until tokens are available.
                      If False, return False immediately when unavailable.

        Returns:
            True if tokens were acquired, False if non-blocking and unavailable.
        """
        while True:
            with self._lock:
                self._refill()
                if self._tokens >= tokens:
                    self._tokens -= tokens
                    return True
                if not blocking:
                    return False
                wait_time = (tokens - self._tokens) / self._rate
            time.sleep(wait_time)

    def _refill(self) -> None:
        """Refill tokens based on elapsed time. Must be called under lock."""
        now = time.monotonic()
        elapsed = now - self._last_refill
        self._tokens = min(self._capacity, self._tokens + elapsed * self._rate)
        self._last_refill = now



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tubefetch.utils.time_fmt — timestamp edge cases."""

from tubefetch.utils.time_fmt import seconds_to_srt, seconds_to_vtt


class TestSecondsToVtt:
    def test_zero(self):
        assert seconds_to_vtt(0.0) == "00:00:00.000"

    def test_milliseconds(self):
        assert seconds_to_vtt(0.5) == "00:00:00.500"

    def test_one_second(self):
        assert seconds_to_vtt(1.0) == "00:00:01.000"

    def test_minutes(self):
        assert seconds_to_vtt(65.0) == "00:01:05.000"

    def test_hours(self):
        assert seconds_to_vtt(3661.5) == "01:01:01.500"

    def test_large_value(self):
        assert seconds_to_vtt(36000.0) == "10:00:00.000"

    def test_fractional_milliseconds(self):
        assert seconds_to_vtt(1.1234) == "00:00:01.123"

    def test_rounding_edge(self):
        assert seconds_to_vtt(1.9999) == "00:00:01.999"

    def test_negative_clamped_to_zero(self):
        assert seconds_to_vtt(-5.0) == "00:00:00.000"

    def test_uses_dot_separator(self):
        result = seconds_to_vtt(1.5)
        assert "." in result
        assert "," not in result

    def test_59_minutes_59_seconds(self):
        assert seconds_to_vtt(3599.999) == "00:59:59.999"

    def test_exact_boundary(self):
        assert seconds_to_vtt(60.0) == "00:01:00.000"

    def test_small_fraction(self):
        assert seconds_to_vtt(0.001) == "00:00:00.001"


class TestSecondsToSrt:
    def test_zero(self):
        assert seconds_to_srt(0.0) == "00:00:00,000"

    def test_milliseconds(self):
        assert seconds_to_srt(0.5) == "00:00:00,500"

    def test_one_second(self):
        assert seconds_to_srt(1.0) == "00:00:01,000"

    def test_minutes(self):
        assert seconds_to_srt(65.0) == "00:01:05,000"

    def test_hours(self):
        assert seconds_to_srt(3661.5) == "01:01:01,500"

    def test_uses_comma_separator(self):
        result = seconds_to_srt(1.5)
        assert "," in result
        # Only the timestamp comma, not a dot
        parts = result.split(",")
        assert len(parts) == 2

    def test_negative_clamped_to_zero(self):
        assert seconds_to_srt(-5.0) == "00:00:00,000"

    def test_rounding_edge(self):
        assert seconds_to_srt(1.9999) == "00:00:01,999"

    def test_small_fraction(self):
        assert seconds_to_srt(0.001) == "00:00:00,001"



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tubefetch.core.logging."""

import json
import logging

from tubefetch.core.logging import (
    JsonlFormatter,
    get_logger,
    log_event,
    setup_logging,
)


class TestSetupLogging:
    def test_returns_logger(self):
        logger = setup_logging()
        assert isinstance(logger, logging.Logger)
        assert logger.name == "tubefetch"

    def test_info_level_by_default(self):
        logger = setup_logging(verbose=False)
        assert logger.level == logging.INFO

    def test_debug_level_when_verbose(self):
        logger = setup_logging(verbose=True)
        assert logger.level == logging.DEBUG

    def test_clears_existing_handlers(self):
        logger = setup_logging()
        count1 = len(logger.handlers)
        setup_logging()
        count2 = len(logger.handlers)
        assert count1 == count2

    def test_jsonl_handler_added(self, tmp_path):
        path = tmp_path / "logs" / "test.jsonl"
        logger = setup_logging(jsonl_path=path)
        handler_types = [type(h).__name__ for h in logger.handlers]
        assert "JsonlFileHandler" in handler_types

    def test_jsonl_file_created(self, tmp_path):
        path = tmp_path / "test.jsonl"
        setup_logging(jsonl_path=path)
        logger = get_logger()
        logger.info("test message")
        assert path.exists()
        content = path.read_text().strip()
        record = json.loads(content)
        assert record["message"] == "test message"
        assert record["level"] == "INFO"


class TestJsonlFormatter:
    def test_format_basic(self):
        formatter = JsonlFormatter()
        record = logging.LogRecord(
            name="yt_fetch",
            level=logging.INFO,
            pathname="",
            lineno=0,
            msg="hello",
            args=(),
            exc_info=None,
        )
        output = formatter.format(record)
        data = json.loads(output)
        assert data["message"] == "hello"
        assert data["level"] == "INFO"
        assert "timestamp" in data

    def test_format_with_extras(self):
        formatter = JsonlFormatter()
        record = logging.LogRecord(
            name="yt_fetch",
            level=logging.ERROR,
            pathname="",
            lineno=0,
            msg="fail",
            args=(),
            exc_info=None,
        )
        record.video_id = "abc12345678"
        record.event = "metadata_fetch"
        record.details = "some detail"
        record.error = "timeout"
        output = formatter.format(record)
        data = json.loads(output)
        assert data["video_id"] == "abc12345678"
        assert data["event"] == "metadata_fetch"
        assert data["details"] == "some detail"
        assert data["error"] == "timeout"


class TestGetLogger:
    def test_returns_same_logger(self):
        setup_logging()
        logger = get_logger()
        assert logger.name == "tubefetch"


class TestLogEvent:
    def test_log_event(self, tmp_path):
        path = tmp_path / "events.jsonl"
        setup_logging(jsonl_path=path)
        log_event(
            logging.INFO,
            "fetched",
            video_id="abc12345678",
            event="metadata_ok",
        )
        content = path.read_text().strip()
        data = json.loads(content)
        assert data["message"] == "fetched"
        assert data["video_id"] == "abc12345678"
        assert data["event"] == "metadata_ok"



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Media download via yt-dlp."""

from __future__ import annotations

import logging
from dataclasses import dataclass, field
from pathlib import Path

import yt_dlp

from tubefetch.core.errors import (
    FetchError,
    FetchErrorCode,
    FetchPhase,
    MediaError,
    MediaServiceError,
    _classify_exception,
)
from tubefetch.core.options import FetchOptions
from tubefetch.utils.ffmpeg import check_ffmpeg

logger = logging.getLogger("tubefetch")

# Retryable error codes for transient failures
RETRYABLE_CODES = (
    FetchErrorCode.NETWORK_ERROR,
    FetchErrorCode.TIMEOUT,
    FetchErrorCode.SERVICE_ERROR,
    FetchErrorCode.RATE_LIMITED,
)


@dataclass
class MediaResult:
    """Result of a media download operation."""

    video_id: str
    paths: list[Path] = field(default_factory=list)
    skipped: bool = False
    errors: list[FetchError] = field(default_factory=list)


def download_media(
    video_id: str,
    options: FetchOptions,
    out_dir: Path,
) -> MediaResult:
    """Download media for a video using yt-dlp.

    Respects options.download mode (none/video/audio/both),
    max_height, format, audio_format, and ffmpeg_fallback.
    """
    if options.download == "none":
        return MediaResult(video_id=video_id, skipped=True)

    has_ffmpeg = check_ffmpeg()
    if not has_ffmpeg:
        if options.ffmpeg_fallback == "skip":
            logger.warning("ffmpeg not found, skipping media download for %s", video_id)
            return MediaResult(
                video_id=video_id,
                skipped=True,
                errors=[
                    FetchError(
                        code=FetchErrorCode.MISSING_DEPENDENCY,
                        message="ffmpeg not found, skipped media download",
                        phase=FetchPhase.MEDIA,
                        retryable=False,
                        video_id=video_id,
                    )
                ],
            )
        else:
            raise MediaError(
                "ffmpeg is required for media download but was not found. "
                "Install ffmpeg or set --ffmpeg-fallback=skip.",
                code=FetchErrorCode.MISSING_DEPENDENCY,
            )

    media_dir = Path(out_dir) / video_id / "media"
    media_dir.mkdir(parents=True, exist_ok=True)

    result = MediaResult(video_id=video_id)
    url = f"https://www.youtube.com/watch?v={video_id}"

    if options.download in ("video", "both"):
        paths = _download_video(url, video_id, options, media_dir)
        result.paths.extend(paths)

    if options.download in ("audio", "both"):
        paths = _download_audio(url, video_id, options, media_dir)
        result.paths.extend(paths)

    return result


def _build_video_format(options: FetchOptions) -> str:
    """Build yt-dlp format string for video download."""
    if options.format != "best":
        return options.format

    if options.max_height:
        return f"bestvideo[height<={options.max_height}]+bestaudio/best[height<={options.max_height}]"

    return "bestvideo+bestaudio/best"


def _build_audio_format(options: FetchOptions) -> str:
    """Build yt-dlp format string for audio-only download."""
    if options.audio_format != "best":
        return f"bestaudio[ext={options.audio_format}]/bestaudio"

    return "bestaudio/best"


def _download_video(
    url: str,
    video_id: str,
    options: FetchOptions,
    media_dir: Path,
) -> list[Path]:
    """Download video via yt-dlp."""
    fmt = _build_video_format(options)
    outtmpl = str(media_dir / f"{video_id}.%(ext)s")

    ydl_opts = {
        "format": fmt,
        "outtmpl": outtmpl,
        "quiet": True,
        "no_warnings": True,
        "no_color": True,
        "merge_output_format": "mp4",
    }

    return _run_yt_dlp(url, video_id, ydl_opts, "video")


def _download_audio(
    url: str,
    video_id: str,
    options: FetchOptions,
    media_dir: Path,
) -> list[Path]:
    """Download audio-only via yt-dlp."""
    fmt = _build_audio_format(options)
    outtmpl = str(media_dir / f"{video_id}_audio.%(ext)s")

    ydl_opts = {
        "format": fmt,
        "outtmpl": outtmpl,
        "quiet": True,
        "no_warnings": True,
        "no_color": True,
        "postprocessors": [
            {
                "key": "FFmpegExtractAudio",
                "preferredcodec": "m4a" if options.audio_format == "best" else options.audio_format,
            }
        ],
    }

    return _run_yt_dlp(url, video_id, ydl_opts, "audio")


def _run_yt_dlp(url: str, video_id: str, ydl_opts: dict, media_type: str) -> list[Path]:
    """Execute yt-dlp download and return list of downloaded file paths."""
    downloaded: list[Path] = []

    def progress_hook(d: dict) -> None:
        if d.get("status") == "finished":
            filename = d.get("filename")
            if filename:
                downloaded.append(Path(filename))

    ydl_opts["progress_hooks"] = [progress_hook]

    try:
        with yt_dlp.YoutubeDL(ydl_opts) as ydl:
            ydl.download([url])
    except yt_dlp.utils.DownloadError as exc:
        code = _classify_exception(exc)
        if code in RETRYABLE_CODES:
            raise MediaServiceError(f"Failed to download {media_type} for {video_id}: {exc}", code=code) from exc
        raise MediaError(f"Failed to download {media_type} for {video_id}: {exc}", code=code) from exc

    return downloaded



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""VTT/SRT timestamp formatting helpers (VTT, SRT)."""

from __future__ import annotations


def seconds_to_vtt(seconds: float) -> str:
    """Convert seconds to WebVTT timestamp: HH:MM:SS.mmm"""
    seconds = max(0.0, seconds)
    h = int(seconds // 3600)
    m = int((seconds % 3600) // 60)
    s = int(seconds % 60)
    ms = int(round((seconds % 1) * 1000))
    if ms >= 1000:
        ms = 999
    return f"{h:02d}:{m:02d}:{s:02d}.{ms:03d}"


def seconds_to_srt(seconds: float) -> str:
    """Convert seconds to SRT timestamp: HH:MM:SS,mmm"""
    seconds = max(0.0, seconds)
    h = int(seconds // 3600)
    m = int((seconds % 3600) // 60)
    s = int(seconds % 60)
    ms = int(round((seconds % 1) * 1000))
    if ms >= 1000:
        ms = 999
    return f"{h:02d}:{m:02d}:{s:02d},{ms:03d}"



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Idempotency tests for yt_fetch.core.pipeline caching behavior.

Verifies that:
- Re-running without --force skips all work (services not called)
- Re-running with --force overwrites all outputs
- Selective --force-metadata, --force-transcript, --force-media work independently
- Cached file paths are returned correctly on skip
"""

from datetime import datetime, timezone
from unittest.mock import patch

from tubefetch.core.models import Metadata, Transcript, TranscriptSegment
from tubefetch.core.options import FetchOptions
from tubefetch.core.pipeline import process_video
from tubefetch.services.media import MediaResult


def _make_metadata(video_id: str = "testVid12345") -> Metadata:
    return Metadata(
        video_id=video_id,
        source_url=f"https://www.youtube.com/watch?v={video_id}",
        title="Test Video",
        fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
        metadata_source="yt-dlp",
    )


def _make_transcript(video_id: str = "testVid12345") -> Transcript:
    return Transcript(
        video_id=video_id,
        language="en",
        is_generated=False,
        segments=[TranscriptSegment(start=0.0, duration=2.0, text="Hello")],
        fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
        transcript_source="youtube-transcript-api",
    )


class TestIdempotencySkip:
    """Re-running without --force skips all work."""

    @patch("tubefetch.core.pipeline.download_media")
    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_second_run_skips_all(self, mock_meta, mock_trans, mock_media, tmp_path):
        # First run: fetch everything
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        opts = FetchOptions(out=tmp_path)

        result1 = process_video("testVid12345", opts)
        assert result1.success is True
        assert result1.metadata_path.exists()
        assert result1.transcript_path.exists()
        assert mock_meta.call_count == 1
        assert mock_trans.call_count == 1

        # Second run: should skip both
        mock_meta.reset_mock()
        mock_trans.reset_mock()

        result2 = process_video("testVid12345", opts)
        assert result2.success is True
        mock_meta.assert_not_called()
        mock_trans.assert_not_called()
        # Cached paths should still be returned
        assert result2.metadata_path == result1.metadata_path
        assert result2.transcript_path == result1.transcript_path

    @patch("tubefetch.core.pipeline.download_media")
    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_second_run_skips_media(self, mock_meta, mock_trans, mock_media, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        mock_media.return_value = MediaResult(
            video_id="testVid12345",
            paths=[tmp_path / "testVid12345" / "media" / "video.mp4"],
        )
        opts = FetchOptions(out=tmp_path, download="video")

        # First run
        process_video("testVid12345", opts)
        assert mock_media.call_count == 1

        # Create a fake media file so cache check passes
        media_dir = tmp_path / "testVid12345" / "media"
        media_dir.mkdir(parents=True, exist_ok=True)
        (media_dir / "video.mp4").write_bytes(b"fake")

        # Second run
        mock_meta.reset_mock()
        mock_trans.reset_mock()
        mock_media.reset_mock()

        result2 = process_video("testVid12345", opts)
        mock_meta.assert_not_called()
        mock_trans.assert_not_called()
        mock_media.assert_not_called()
        # Cached media paths returned
        assert len(result2.media_paths) == 1

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_skipped_result_is_still_success(self, mock_meta, mock_trans, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        opts = FetchOptions(out=tmp_path)

        process_video("testVid12345", opts)
        mock_meta.reset_mock()
        mock_trans.reset_mock()

        result = process_video("testVid12345", opts)
        assert result.success is True
        assert result.errors == []


class TestIdempotencyForce:
    """Re-running with --force overwrites all outputs."""

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_force_refetches_all(self, mock_meta, mock_trans, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        opts = FetchOptions(out=tmp_path)

        # First run
        process_video("testVid12345", opts)

        # Second run with --force
        mock_meta.reset_mock()
        mock_trans.reset_mock()
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()

        opts_force = FetchOptions(out=tmp_path, force=True)
        result = process_video("testVid12345", opts_force)

        mock_meta.assert_called_once()
        mock_trans.assert_called_once()
        assert result.success is True

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_force_overwrites_files(self, mock_meta, mock_trans, tmp_path):
        # First run with title "Original"
        meta1 = _make_metadata()
        meta1.title = "Original"
        mock_meta.return_value = meta1
        mock_trans.return_value = _make_transcript()
        opts = FetchOptions(out=tmp_path)
        process_video("testVid12345", opts)

        import json

        data1 = json.loads((tmp_path / "testVid12345" / "metadata.json").read_text())
        assert data1["title"] == "Original"

        # Second run with --force and title "Updated"
        meta2 = _make_metadata()
        meta2.title = "Updated"
        mock_meta.return_value = meta2
        mock_trans.return_value = _make_transcript()

        opts_force = FetchOptions(out=tmp_path, force=True)
        process_video("testVid12345", opts_force)

        data2 = json.loads((tmp_path / "testVid12345" / "metadata.json").read_text())
        assert data2["title"] == "Updated"


class TestSelectiveForce:
    """Selective --force-metadata, --force-transcript, --force-media."""

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_force_metadata_only(self, mock_meta, mock_trans, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        opts = FetchOptions(out=tmp_path)
        process_video("testVid12345", opts)

        mock_meta.reset_mock()
        mock_trans.reset_mock()
        mock_meta.return_value = _make_metadata()

        opts_fm = FetchOptions(out=tmp_path, force_metadata=True)
        process_video("testVid12345", opts_fm)

        mock_meta.assert_called_once()
        mock_trans.assert_not_called()

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_force_transcript_only(self, mock_meta, mock_trans, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        opts = FetchOptions(out=tmp_path)
        process_video("testVid12345", opts)

        mock_meta.reset_mock()
        mock_trans.reset_mock()
        mock_trans.return_value = _make_transcript()

        opts_ft = FetchOptions(out=tmp_path, force_transcript=True)
        process_video("testVid12345", opts_ft)

        mock_meta.assert_not_called()
        mock_trans.assert_called_once()

    @patch("tubefetch.core.pipeline.download_media")
    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_force_media_only(self, mock_meta, mock_trans, mock_media, tmp_path):
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        mock_media.return_value = MediaResult(video_id="testVid12345")
        opts = FetchOptions(out=tmp_path, download="video")
        process_video("testVid12345", opts)

        # Create cached media
        media_dir = tmp_path / "testVid12345" / "media"
        media_dir.mkdir(parents=True, exist_ok=True)
        (media_dir / "video.mp4").write_bytes(b"fake")

        mock_meta.reset_mock()
        mock_trans.reset_mock()
        mock_media.reset_mock()
        mock_media.return_value = MediaResult(video_id="testVid12345")

        opts_fmedia = FetchOptions(out=tmp_path, download="video", force_media=True)
        process_video("testVid12345", opts_fmedia)

        mock_meta.assert_not_called()
        mock_trans.assert_not_called()
        mock_media.assert_called_once()

    @patch("tubefetch.core.pipeline.get_transcript")
    @patch("tubefetch.core.pipeline.get_metadata")
    def test_force_overrides_selective(self, mock_meta, mock_trans, tmp_path):
        """--force should override even if selective flags are not set."""
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()
        opts = FetchOptions(out=tmp_path)
        process_video("testVid12345", opts)

        mock_meta.reset_mock()
        mock_trans.reset_mock()
        mock_meta.return_value = _make_metadata()
        mock_trans.return_value = _make_transcript()

        opts_force = FetchOptions(out=tmp_path, force=True)
        process_video("testVid12345", opts_force)

        mock_meta.assert_called_once()
        mock_trans.assert_called_once()



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for gentlify throttle configuration and retry logic."""

import pytest

from tubefetch.core.errors import (
    FetchErrorCode,
    MetadataError,
    MetadataServiceError,
    TranscriptError,
    TranscriptServiceError,
)
from tubefetch.core.options import FetchOptions
from tubefetch.utils.gentlify_config import (
    _is_retryable_exception,
    async_execute_with_retry,
    create_throttle,
    execute_with_retry,
)


class TestCreateThrottle:
    """Test throttle creation with different FetchOptions configurations."""

    def test_creates_throttle_with_retry_enabled(self):
        """Throttle should be created with retry config when retries > 0."""
        options = FetchOptions(retries=3)
        throttle = create_throttle(options)

        assert throttle is not None
        # Throttle is created successfully (internal retry config not exposed)

    def test_creates_throttle_with_retry_disabled(self):
        """Throttle should be created without retry config when retries = 0."""
        options = FetchOptions(retries=0)
        throttle = create_throttle(options)

        assert throttle is not None
        # Throttle is created successfully (internal retry config not exposed)

    def test_default_retry_count(self):
        """Default retry count should be 3."""
        options = FetchOptions()
        throttle = create_throttle(options)

        assert throttle is not None
        # Default FetchOptions.retries is 3


class TestIsRetryableException:
    """Test exception retryability classification."""

    def test_fetch_exception_with_retryable_true(self):
        """FetchException with retryable=True should be retryable."""
        exc = MetadataServiceError("transient error")
        assert exc.retryable is True
        assert _is_retryable_exception(exc) is True

    def test_fetch_exception_with_retryable_false(self):
        """FetchException with retryable=False should not be retryable."""
        exc = MetadataError("permanent error", code=FetchErrorCode.VIDEO_NOT_FOUND)
        assert exc.retryable is False
        assert _is_retryable_exception(exc) is False

    def test_connection_error_is_retryable(self):
        """ConnectionError should be retryable (fallback)."""
        exc = ConnectionError("network error")
        assert _is_retryable_exception(exc) is True

    def test_timeout_error_is_retryable(self):
        """TimeoutError should be retryable (fallback)."""
        exc = TimeoutError("timeout")
        assert _is_retryable_exception(exc) is True

    def test_os_error_is_retryable(self):
        """OSError should be retryable (fallback)."""
        exc = OSError("os error")
        assert _is_retryable_exception(exc) is True

    def test_generic_exception_not_retryable(self):
        """Generic exceptions should not be retryable."""
        exc = ValueError("bad value")
        assert _is_retryable_exception(exc) is False

    def test_transcript_service_error_retryable(self):
        """TranscriptServiceError should be retryable."""
        exc = TranscriptServiceError("service error")
        assert exc.retryable is True
        assert _is_retryable_exception(exc) is True

    def test_transcript_error_not_retryable(self):
        """TranscriptError (base) should not be retryable."""
        exc = TranscriptError("permanent error", code=FetchErrorCode.TRANSCRIPT_NOT_FOUND)
        assert exc.retryable is False
        assert _is_retryable_exception(exc) is False


class TestExecuteWithRetry:
    """Test retry execution wrapper."""

    @pytest.mark.asyncio
    async def test_async_execute_success_first_attempt(self):
        """Function should succeed on first attempt without retry."""
        call_count = 0

        def successful_func(arg1, arg2):
            nonlocal call_count
            call_count += 1
            return f"{arg1}-{arg2}"

        options = FetchOptions(retries=3)
        throttle = create_throttle(options)

        result = await async_execute_with_retry(successful_func, throttle, "foo", "bar")

        assert result == "foo-bar"
        assert call_count == 1

    @pytest.mark.asyncio
    async def test_async_execute_retries_on_retryable_error(self):
        """Function should retry on retryable errors."""
        call_count = 0

        def flaky_func():
            nonlocal call_count
            call_count += 1
            if call_count < 3:
                raise MetadataServiceError("transient")
            return "success"

        options = FetchOptions(retries=3)
        throttle = create_throttle(options)

        result = await async_execute_with_retry(flaky_func, throttle)

        assert result == "success"
        assert call_count == 3

    @pytest.mark.asyncio
    async def test_async_execute_fails_on_non_retryable_error(self):
        """Function should not retry on non-retryable errors."""
        call_count = 0

        def permanent_fail():
            nonlocal call_count
            call_count += 1
            raise MetadataError("permanent", code=FetchErrorCode.VIDEO_NOT_FOUND)

        options = FetchOptions(retries=3)
        throttle = create_throttle(options)

        with pytest.raises(MetadataError, match="permanent"):
            await async_execute_with_retry(permanent_fail, throttle)

        assert call_count == 1  # No retries for non-retryable errors

    @pytest.mark.asyncio
    async def test_async_execute_exhausts_retries(self):
        """Function should raise after exhausting all retries."""
        call_count = 0

        def always_fails():
            nonlocal call_count
            call_count += 1
            raise ConnectionError("always fails")

        options = FetchOptions(retries=2)
        throttle = create_throttle(options)

        with pytest.raises(ConnectionError, match="always fails"):
            await async_execute_with_retry(always_fails, throttle)

        assert call_count == 2  # max_attempts = 2

    @pytest.mark.asyncio
    async def test_async_execute_no_retry_when_disabled(self):
        """Function should not retry when retries=0."""
        call_count = 0

        def flaky_func():
            nonlocal call_count
            call_count += 1
            raise ConnectionError("error")

        options = FetchOptions(retries=0)
        throttle = create_throttle(options)

        with pytest.raises(ConnectionError, match="error"):
            await async_execute_with_retry(flaky_func, throttle)

        assert call_count == 1  # No retries

    def test_sync_execute_with_retry(self):
        """Sync wrapper should work correctly."""
        call_count = 0

        def flaky_func(value):
            nonlocal call_count
            call_count += 1
            if call_count < 2:
                raise TimeoutError("timeout")
            return value * 2

        options = FetchOptions(retries=3)
        throttle = create_throttle(options)

        result = execute_with_retry(flaky_func, throttle, 5)

        assert result == 10
        assert call_count == 2

    def test_sync_execute_with_kwargs(self):
        """Sync wrapper should handle kwargs correctly."""

        def func_with_kwargs(a, b=10):
            return a + b

        options = FetchOptions(retries=1)
        throttle = create_throttle(options)

        result = execute_with_retry(func_with_kwargs, throttle, 5, b=20)

        assert result == 25


class TestRetryBackoff:
    """Test retry backoff behavior."""

    @pytest.mark.asyncio
    async def test_exponential_jitter_backoff(self):
        """Retry should use exponential jitter backoff."""
        call_count = 0

        def flaky_func():
            nonlocal call_count
            call_count += 1
            if call_count < 3:
                raise ConnectionError("transient")
            return "success"

        options = FetchOptions(retries=3)
        throttle = create_throttle(options)

        result = await async_execute_with_retry(flaky_func, throttle)

        assert result == "success"
        assert call_count == 3  # Retried until success


class TestIntegrationWithPipeline:
    """Test gentlify integration with actual pipeline usage patterns."""

    def test_retries_respect_fetch_options(self):
        """Retry count should respect FetchOptions.retries setting."""
        call_count = 0

        def flaky_metadata(video_id, options):
            nonlocal call_count
            call_count += 1
            if call_count < 2:
                raise MetadataServiceError("transient")
            return {"video_id": video_id}

        # Test with retries enabled
        options = FetchOptions(retries=3)
        throttle = create_throttle(options)
        result = execute_with_retry(flaky_metadata, throttle, "test123", options)

        assert result["video_id"] == "test123"
        assert call_count == 2

    def test_zero_retries_fails_immediately(self):
        """With retries=0, errors should propagate immediately."""
        call_count = 0

        def always_fails(video_id):
            nonlocal call_count
            call_count += 1
            raise MetadataServiceError("error")

        options = FetchOptions(retries=0)
        throttle = create_throttle(options)

        with pytest.raises(MetadataServiceError):
            execute_with_retry(always_fails, throttle, "test123")

        assert call_count == 1  # Should fail immediately, no retries



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Structured logging setup (console + JSONL)."""

from __future__ import annotations

import json
import logging
from datetime import datetime, timezone
from pathlib import Path

from rich.console import Console
from rich.logging import RichHandler

_console = Console(stderr=True)


class JsonlFormatter(logging.Formatter):
    """Formats log records as single-line JSON objects."""

    def format(self, record: logging.LogRecord) -> str:
        entry: dict = {
            "timestamp": datetime.now(timezone.utc).isoformat(),
            "level": record.levelname,
            "video_id": getattr(record, "video_id", None),
            "event": getattr(record, "event", None),
            "details": getattr(record, "details", None),
            "error": getattr(record, "error", None),
        }
        if record.getMessage():
            entry["message"] = record.getMessage()
        return json.dumps(entry, default=str)


class JsonlFileHandler(logging.FileHandler):
    """File handler that uses JsonlFormatter by default."""

    def __init__(self, path: Path) -> None:
        path.parent.mkdir(parents=True, exist_ok=True)
        super().__init__(str(path), mode="a", encoding="utf-8")
        self.setFormatter(JsonlFormatter())


def setup_logging(*, verbose: bool = False, jsonl_path: Path | None = None) -> logging.Logger:
    """Configure and return the yt_fetch logger.

    Args:
        verbose: If True, set level to DEBUG; otherwise INFO.
        jsonl_path: If provided, also write structured JSONL logs to this file.

    Returns:
        The configured 'yt_fetch' logger.
    """
    logger = logging.getLogger("tubefetch")
    logger.handlers.clear()
    logger.setLevel(logging.DEBUG if verbose else logging.INFO)

    rich_handler = RichHandler(
        console=_console,
        show_time=verbose,
        show_path=verbose,
        rich_tracebacks=True,
        markup=True,
    )
    rich_handler.setLevel(logging.DEBUG if verbose else logging.INFO)
    logger.addHandler(rich_handler)

    if jsonl_path is not None:
        jsonl_handler = JsonlFileHandler(jsonl_path)
        jsonl_handler.setLevel(logging.DEBUG)
        logger.addHandler(jsonl_handler)

    return logger


def get_logger() -> logging.Logger:
    """Get the yt_fetch logger (must call setup_logging first)."""
    return logging.getLogger("tubefetch")


def log_event(
    level: int,
    message: str,
    *,
    video_id: str | None = None,
    event: str | None = None,
    details: str | None = None,
    error: str | None = None,
) -> None:
    """Log a structured event with optional yt-fetch-specific fields."""
    logger = get_logger()
    logger.log(
        level,
        message,
        extra={
            "video_id": video_id,
            "event": event,
            "details": details,
            "error": error,
        },
    )



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""yt-fetch — YouTube video metadata, transcript, and media fetcher."""

__version__ = "0.8.0"

from tubefetch.core.errors import FetchError, FetchErrorCode, FetchException, FetchPhase
from tubefetch.core.models import BatchResult, FetchResult, Metadata, Transcript
from tubefetch.core.options import FetchOptions


def fetch_video(video_id: str, options: FetchOptions | None = None) -> FetchResult:
    """Fetch metadata, transcript, and optionally media for a single video.

    This is the primary library entry point for single-video processing.

    Args:
        video_id: YouTube video ID or URL.
        options: Configuration options. Uses defaults if not provided.

    Returns:
        FetchResult with metadata, transcript, paths, and any errors.
    """
    from tubefetch.core.pipeline import process_video
    from tubefetch.services.id_parser import parse_video_id

    if options is None:
        options = FetchOptions()

    parsed = parse_video_id(video_id)
    if parsed is None:
        return FetchResult(
            video_id=video_id,
            success=False,
            errors=[
                FetchError(
                    code=FetchErrorCode.INVALID_VIDEO_ID,
                    message=f"Invalid video ID or URL: {video_id}",
                    phase=FetchPhase.METADATA,
                    retryable=False,
                    video_id=video_id,
                )
            ],
        )

    return process_video(parsed, options)


def fetch_batch(video_ids: list[str], options: FetchOptions | None = None) -> BatchResult:
    """Fetch metadata, transcript, and optionally media for multiple videos.

    This is the primary library entry point for batch processing.

    Args:
        video_ids: List of YouTube video IDs or URLs.
        options: Configuration options. Uses defaults if not provided.

    Returns:
        BatchResult with per-video results and summary counts.
    """
    from tubefetch.core.pipeline import process_batch
    from tubefetch.services.id_parser import parse_many

    if options is None:
        options = FetchOptions()

    parsed = parse_many(video_ids)
    return process_batch(parsed, options)


__all__ = [
    "__version__",
    "fetch_video",
    "fetch_batch",
    "FetchOptions",
    "FetchResult",
    "BatchResult",
    "Metadata",
    "Transcript",
    "FetchError",
    "FetchErrorCode",
    "FetchPhase",
    "FetchException",
]



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tubefetch.core.models and yt_fetch.core.options."""

from datetime import datetime, timezone
from pathlib import Path

import pytest
from pydantic import ValidationError

from tubefetch.core.errors import FetchError, FetchErrorCode, FetchPhase
from tubefetch.core.models import (
    BatchResult,
    FetchResult,
    Metadata,
    Transcript,
    TranscriptSegment,
)
from tubefetch.core.options import FetchOptions

# --- Metadata ---


class TestMetadata:
    def test_minimal(self):
        m = Metadata(
            video_id="dQw4w9WgXcQ",
            source_url="https://www.youtube.com/watch?v=dQw4w9WgXcQ",
            fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
            metadata_source="yt-dlp",
        )
        assert m.video_id == "dQw4w9WgXcQ"
        assert m.title is None
        assert m.tags == []
        assert m.raw is None

    def test_full(self):
        m = Metadata(
            video_id="dQw4w9WgXcQ",
            source_url="https://www.youtube.com/watch?v=dQw4w9WgXcQ",
            title="Never Gonna Give You Up",
            channel_title="Rick Astley",
            channel_id="UCuAXFkgsw1L7xaCfnd5JJOw",
            upload_date="2009-10-25",
            duration_seconds=212.0,
            description="The official video",
            tags=["rick", "astley"],
            view_count=1_500_000_000,
            like_count=15_000_000,
            fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
            metadata_source="yt-dlp",
            raw={"id": "dQw4w9WgXcQ"},
        )
        assert m.title == "Never Gonna Give You Up"
        assert m.duration_seconds == 212.0
        assert len(m.tags) == 2

    def test_missing_required_fields(self):
        with pytest.raises(ValidationError):
            Metadata()

    def test_round_trip(self):
        m = Metadata(
            video_id="abc",
            source_url="https://youtu.be/abc",
            fetched_at=datetime(2025, 6, 1, tzinfo=timezone.utc),
            metadata_source="yt-dlp",
        )
        data = m.model_dump()
        m2 = Metadata.model_validate(data)
        assert m == m2

    def test_json_round_trip(self):
        m = Metadata(
            video_id="abc",
            source_url="https://youtu.be/abc",
            fetched_at=datetime(2025, 6, 1, tzinfo=timezone.utc),
            metadata_source="yt-dlp",
        )
        json_str = m.model_dump_json()
        m2 = Metadata.model_validate_json(json_str)
        assert m == m2


# --- TranscriptSegment ---


class TestTranscriptSegment:
    def test_valid(self):
        seg = TranscriptSegment(start=0.0, duration=5.5, text="Hello world")
        assert seg.start == 0.0
        assert seg.duration == 5.5
        assert seg.text == "Hello world"

    def test_missing_text(self):
        with pytest.raises(ValidationError):
            TranscriptSegment(start=0.0, duration=1.0)


# --- Transcript ---


class TestTranscript:
    def test_minimal(self):
        t = Transcript(
            video_id="abc",
            language="en",
            segments=[],
            fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
            transcript_source="youtube-transcript-api",
        )
        assert t.is_generated is None
        assert t.available_languages == []
        assert t.errors == []

    def test_with_segments(self):
        t = Transcript(
            video_id="abc",
            language="en",
            is_generated=False,
            segments=[
                TranscriptSegment(start=0.0, duration=2.0, text="Hello"),
                TranscriptSegment(start=2.0, duration=3.0, text="World"),
            ],
            fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
            transcript_source="youtube-transcript-api",
            available_languages=["en", "es"],
        )
        assert len(t.segments) == 2
        assert t.segments[0].text == "Hello"
        assert t.available_languages == ["en", "es"]

    def test_round_trip(self):
        t = Transcript(
            video_id="abc",
            language="en",
            segments=[TranscriptSegment(start=0.0, duration=1.0, text="Hi")],
            fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
            transcript_source="youtube-transcript-api",
        )
        data = t.model_dump()
        t2 = Transcript.model_validate(data)
        assert t == t2


# --- FetchResult ---


class TestFetchResult:
    def test_success(self):
        r = FetchResult(
            video_id="abc",
            success=True,
            metadata_path=Path("out/abc/metadata.json"),
        )
        assert r.success is True
        assert r.errors == []
        assert r.media_paths == []

    def test_failure(self):
        r = FetchResult(
            video_id="abc",
            success=False,
            errors=[
                FetchError(
                    code=FetchErrorCode.VIDEO_NOT_FOUND,
                    message="Video not found",
                    phase=FetchPhase.METADATA,
                    retryable=False,
                    video_id="abc",
                )
            ],
        )
        assert r.success is False
        assert len(r.errors) == 1

    def test_round_trip(self):
        r = FetchResult(video_id="abc", success=True)
        data = r.model_dump()
        r2 = FetchResult.model_validate(data)
        assert r == r2


# --- BatchResult ---


class TestBatchResult:
    def test_valid(self):
        b = BatchResult(
            total=3,
            succeeded=2,
            failed=1,
            results=[
                FetchResult(video_id="a", success=True),
                FetchResult(video_id="b", success=True),
                FetchResult(
                    video_id="c",
                    success=False,
                    errors=[
                        FetchError(
                            code=FetchErrorCode.UNKNOWN,
                            message="fail",
                            phase=FetchPhase.METADATA,
                            retryable=False,
                            video_id="c",
                        )
                    ],
                ),
            ],
        )
        assert b.total == 3
        assert b.succeeded == 2
        assert b.failed == 1
        assert len(b.results) == 3

    def test_missing_results(self):
        with pytest.raises(ValidationError):
            BatchResult(total=0, succeeded=0, failed=0)


# --- FetchOptions ---


class TestFetchOptions:
    def test_defaults(self):
        opts = FetchOptions()
        assert opts.out == Path("./out")
        assert opts.languages == ["en"]
        assert opts.allow_generated is True
        assert opts.allow_any_language is False
        assert opts.download == "none"
        assert opts.max_height is None
        assert opts.format == "best"
        assert opts.audio_format == "best"
        assert opts.force is False
        assert opts.force_metadata is False
        assert opts.force_transcript is False
        assert opts.force_media is False
        assert opts.retries == 3
        assert opts.rate_limit == 2.0
        assert opts.workers == 3
        assert opts.fail_fast is False
        assert opts.verbose is False
        assert opts.yt_api_key is None
        assert opts.ffmpeg_fallback == "error"

    def test_env_override(self, monkeypatch):
        monkeypatch.setenv("TUBEFETCH_RETRIES", "5")
        opts = FetchOptions()
        assert opts.retries == 5
        monkeypatch.setenv("TUBEFETCH_VERBOSE", "true")
        opts = FetchOptions()
        assert opts.verbose is True
        monkeypatch.setenv("TUBEFETCH_DOWNLOAD", "audio")
        opts = FetchOptions()
        assert opts.download == "audio"

    def test_invalid_download_value(self):
        with pytest.raises(ValidationError):
            FetchOptions(download="invalid")

    def test_invalid_ffmpeg_fallback(self):
        with pytest.raises(ValidationError):
            FetchOptions(ffmpeg_fallback="invalid")

    def test_explicit_values(self):
        opts = FetchOptions(
            out=Path("/tmp/test"),
            languages=["es", "en"],
            workers=8,
            yt_api_key="test-key-123",
        )
        assert opts.out == Path("/tmp/test")
        assert opts.languages == ["es", "en"]
        assert opts.workers == 8
        assert opts.yt_api_key == "test-key-123"



# Copyright (c) 2026 Pointmatic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tubefetch.core.writer."""

import json
from datetime import datetime, timezone

from tubefetch.core.errors import FetchError, FetchErrorCode, FetchPhase
from tubefetch.core.models import (
    BatchResult,
    FetchResult,
    Metadata,
    Transcript,
    TranscriptSegment,
)
from tubefetch.core.writer import (
    read_metadata,
    read_transcript_json,
    write_metadata,
    write_summary,
    write_transcript_json,
    write_transcript_srt,
    write_transcript_txt,
    write_transcript_vtt,
)


def _make_metadata() -> Metadata:
    return Metadata(
        video_id="dQw4w9WgXcQ",
        source_url="https://www.youtube.com/watch?v=dQw4w9WgXcQ",
        title="Test Video",
        channel_title="Test Channel",
        upload_date="2025-01-01",
        duration_seconds=120.0,
        tags=["test"],
        fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
        metadata_source="yt-dlp",
    )


def _make_transcript() -> Transcript:
    return Transcript(
        video_id="dQw4w9WgXcQ",
        language="en",
        is_generated=False,
        segments=[
            TranscriptSegment(start=0.0, duration=2.5, text="Hello world"),
            TranscriptSegment(start=2.5, duration=3.0, text="This is a test"),
            TranscriptSegment(start=5.5, duration=1.5, text="Goodbye"),
        ],
        fetched_at=datetime(2025, 1, 1, tzinfo=timezone.utc),
        transcript_source="youtube-transcript-api",
    )


# --- write_metadata ---


class TestWriteMetadata:
    def test_writes_json(self, tmp_path):
        meta = _make_metadata()
        path = write_metadata(meta, tmp_path)
        assert path.exists()
        assert path.name == "metadata.json"
        assert path.parent.name == "dQw4w9WgXcQ"

    def test_json_structure(self, tmp_path):
        meta = _make_metadata()
        path = write_metadata(meta, tmp_path)
        data = json.loads(path.read_text())
        assert data["video_id"] == "dQw4w9WgXcQ"
        assert data["title"] == "Test Video"
        assert data["metadata_source"] == "yt-dlp"
        assert data["tags"] == ["test"]

    def test_overwrites_existing(self, tmp_path):
        meta = _make_metadata()
        write_metadata(meta, tmp_path)
        meta2 = _make_metadata()
        meta2.title = "Updated"
        path = write_metadata(meta2, tmp_path)
        data = json.loads(path.read_text())
        assert data["title"] == "Updated"


# --- read_metadata ---


class TestReadMetadata:
    def test_round_trip(self, tmp_path):
        meta = _make_metadata()
        write_metadata(meta, tmp_path)
        loaded = read_metadata(tmp_path, "dQw4w9WgXcQ")
        assert loaded is not None
        assert loaded.video_id == "dQw4w9WgXcQ"
        assert loaded.title == "Test Video"
        assert loaded.metadata_source == "yt-dlp"
        assert loaded.tags == ["test"]

    def test_missing_file_returns_none(self, tmp_path):
        loaded = read_metadata(tmp_path, "nonexistent")
        assert loaded is None

    def test_invalid_json_returns_none(self, tmp_path):
        video_dir = tmp_path / "bad"
        video_dir.mkdir()
        (video_dir / "metadata.json").write_text("not json")
        loaded = read_metadata(tmp_path, "bad")
        assert loaded is None

    def test_incomplete_json_returns_none(self, tmp_path):
        video_dir = tmp_path / "bad2"
        video_dir.mkdir()
        (video_dir / "metadata.json").write_text('{"video_id": "bad2"}')
        loaded = read_metadata(tmp_path, "bad2")
        assert loaded is None


# --- read_transcript_json ---


class TestReadTranscriptJson:
    def test_round_trip(self, tmp_path):
        trans = _make_transcript()
        write_transcript_json(trans, tmp_path)
        loaded = read_transcript_json(tmp_path, "dQw4w9WgXcQ")
        assert loaded is not None
        assert loaded.video_id == "dQw4w9WgXcQ"
        assert loaded.language == "en"
        assert len(loaded.segments) == 3
        assert loaded.segments[0].text == "Hello world"

    def test_missing_file_returns_none(self, tmp_path):
        loaded = read_transcript_json(tmp_path, "nonexistent")
        assert loaded is None

    def test_invalid_json_returns_none(self, tmp_path):
        video_dir = tmp_path / "bad"
        video_dir.mkdir()
        (video_dir / "transcript.json").write_text("not json")
        loaded = read_transcript_json(tmp_path, "bad")
        assert loaded is None


# --- write_transcript_json ---


class TestWriteTranscriptJson:
    def test_writes_json(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_json(t, tmp_path)
        assert path.exists()
        assert path.name == "transcript.json"

    def test_json_structure(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_json(t, tmp_path)
        data = json.loads(path.read_text())
        assert data["video_id"] == "dQw4w9WgXcQ"
        assert data["language"] == "en"
        assert len(data["segments"]) == 3
        assert data["segments"][0]["text"] == "Hello world"
        assert data["segments"][0]["start"] == 0.0
        assert data["segments"][0]["duration"] == 2.5


# --- write_transcript_txt ---


class TestWriteTranscriptTxt:
    def test_writes_txt(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_txt(t, tmp_path)
        assert path.exists()
        assert path.name == "transcript.txt"

    def test_no_timestamps(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_txt(t, tmp_path)
        content = path.read_text()
        assert "00:" not in content
        assert "-->" not in content

    def test_plain_text_content(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_txt(t, tmp_path)
        lines = path.read_text().strip().split("\n")
        assert lines == ["Hello world", "This is a test", "Goodbye"]


# --- write_transcript_vtt ---


class TestWriteTranscriptVtt:
    def test_writes_vtt(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_vtt(t, tmp_path)
        assert path.exists()
        assert path.name == "transcript.vtt"

    def test_vtt_header(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_vtt(t, tmp_path)
        content = path.read_text()
        assert content.startswith("WEBVTT\n")

    def test_vtt_timestamps(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_vtt(t, tmp_path)
        content = path.read_text()
        assert "00:00:00.000 --> 00:00:02.500" in content
        assert "00:00:02.500 --> 00:00:05.500" in content
        assert "00:00:05.500 --> 00:00:07.000" in content

    def test_vtt_uses_dot_separator(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_vtt(t, tmp_path)
        content = path.read_text()
        assert "." in content
        assert "," not in content.replace("WEBVTT", "")


# --- write_transcript_srt ---


class TestWriteTranscriptSrt:
    def test_writes_srt(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_srt(t, tmp_path)
        assert path.exists()
        assert path.name == "transcript.srt"

    def test_srt_sequence_numbers(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_srt(t, tmp_path)
        content = path.read_text()
        lines = content.strip().split("\n")
        assert lines[0] == "1"
        assert lines[4] == "2"
        assert lines[8] == "3"

    def test_srt_timestamps(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_srt(t, tmp_path)
        content = path.read_text()
        assert "00:00:00,000 --> 00:00:02,500" in content
        assert "00:00:02,500 --> 00:00:05,500" in content

    def test_srt_uses_comma_separator(self, tmp_path):
        t = _make_transcript()
        path = write_transcript_srt(t, tmp_path)
        content = path.read_text()
        # SRT uses comma in timestamps, not dot
        assert "00:00:00,000" in content


# --- write_summary ---


class TestWriteSummary:
    def test_writes_summary(self, tmp_path):
        batch = BatchResult(
            total=2,
            succeeded=1,
            failed=1,
            results=[
                FetchResult(video_id="a", success=True),
                FetchResult(
                    video_id="b",
                    success=False,
                    errors=[
                        FetchError(
                            code=FetchErrorCode.UNKNOWN,
                            message="fail",
                            phase=FetchPhase.METADATA,
                            retryable=False,
                            video_id="b",
                        )
                    ],
                ),
            ],
        )
        path = write_summary(batch, tmp_path)
        assert path.exists()
        assert path.name == "summary.json"

    def test_summary_structure(self, tmp_path):
        batch = BatchResult(
            total=1,
            succeeded=1,
            failed=0,
            results=[FetchResult(video_id="a", success=True)],
        )
        path = write_summary(batch, tmp_path)
        data = json.loads(path.read_text())
        assert data["total"] == 1
        assert data["succeeded"] == 1
        assert data["failed"] == 0
        assert len(data["results"]) == 1
